"""
Simple entrypoint
"""

import argparse as ap
from pathlib import Path

from lmsi.core import core

parser = ap.ArgumentParser(description="Create a webpage from a set of plots.")

parser.add_argument(
    "--files",
    type=Path,
    nargs="+",
    help="Filenames of the plots to include in the webpage.",
)

parser.add_argument(
    "--config",
    type=Path,
    help="The configuration file to use.",
    default=None,
)

parser.add_argument(
    "--output",
    type=Path,
    help="The output HTML file.",
)

parser.add_argument(
    "--debug",
    action="store_true",
    help="Print debug information.",
    default=False,
)

args = parser.parse_args()


def main():
    if args.debug:
        print(args)

    core(args.files, args.config, args.output)



"""
Functions that aid in the production of the HTML webpages.
"""

from importlib.metadata import PackageNotFoundError, version
from time import strftime

import unyt
from jinja2 import Environment, PackageLoader, select_autoescape

from lmsi.container import PlotContainer


def format_number(number):
    """
    Formats a number from float (with or without units) to a latex-like number.
    """

    try:
        units = f"\\; {number.units.latex_repr}"
    except AttributeError:
        units = ""

    try:
        mantissa, exponent = ("%.3g" % number).split("e+")
        exponent = f" \\times 10^{{{int(exponent)}}}"
    except ValueError:
        mantissa = "%.3g" % number
        exponent = ""

    return f"\\({mantissa}{exponent}{units}\\)"


def get_if_present_float(dictionary, value: str, input_unit=None, output_unit=None):
    """
    A replacement for .get() that also formats the number if present.
    Assumes data should be a float.
    """

    try:
        value = float(dictionary[value])

        if input_unit is not None:
            value = unyt.unyt_quantity(value, input_unit)

            if output_unit is not None:
                value.convert_to_units(output_unit)

        return format_number(value)
    except KeyError:
        return ""


def get_if_present_int(dictionary, value: str, input_unit=None, output_unit=None):
    """
    A replacement for .get() that also formats the number if present.
    Assumes data should be an integer.
    """

    try:
        value = int(dictionary[value])

        if input_unit is not None:
            value = unyt.unyt_quantity(value, input_unit)

            if output_unit is not None:
                value.convert_to_units(output_unit)

        return format_number(value)
    except KeyError:
        return ""


def camel_to_title(string):
    return string.title().replace("_", " ")


class WebpageCreator(object):
    """
    Creates webpages based on the information that is provided in
    the plots metadata through the autoplotter and the additional
    plotting interface provided through the pipeline.
    """

    environment: Environment
    loader: PackageLoader

    variables: dict
    html: str

    plot_container: PlotContainer

    def __init__(self):
        """
        Sets up the ``jinja`` templating system.
        """

        self.loader = PackageLoader("lmsi", "templates")
        self.environment = Environment(
            loader=self.loader, autoescape=select_autoescape(["js"])
        )

        # Initialise empty variables dictionary, with the versions of
        # this package and the velociraptor package used.
        try:
            software_version = version("let-me-scroll-it")
        except PackageNotFoundError:
            # Minor packaging problem who cares...
            software_version = "unknown"

        self.variables = dict(
            creation_date=strftime(r"%Y-%m-%d"),
            pipeline_version=software_version,
            sections={},
            runs=[],
        )

        return

    def render_webpage(self, template: str = "plot_viewer.html") -> str:
        """
        Renders a webpage based on the internal variables stored in
        the ``variables`` dictionary.
        Parameters
        ----------
        template: str
            The name of the template that you wish to use. Defaults to
            "plot_viewer.html".
        Returns
        -------
        html: str
            The resulting HTML. This is also stored in ``.html``.
        """

        self.html = self.environment.get_template(template, parent="base.html").render(
            **self.variables
        )

        return self.html

    def add_metadata(self, page_name: str):
        """
        Add additional metadata to the page.
        Parameters
        ----------
        page_name: str
            Name to put in the page title.
        """

        self.variables.update(dict(page_name=page_name))

    def add_plots(self, plot_container: PlotContainer):
        """
        Adds the auto plotter metadata to the section / plot metadata.
        Parameters
        ----------
        plot_container: PlotContainer
            Complete plot container, post-run.
        """

        self.plot_container = plot_container
        self.variables = {**self.variables, **self.plot_container.model_dump()}

        return

    def save_html(self, filename: str):
        """
        Saves the html in ``self.html`` to the filename provided.
        Parameters
        ----------
        filename: str
            Full filename (including file path) to save the HTML as.
        """

        with open(filename, "w") as handle:
            handle.write(self.html)



"""
The plot container!
"""

import json
import re

from pydantic import BaseModel


class PlotConfig(BaseModel):
    source: str
    title: str
    caption: str
    regex: bool = False

    def match(self, filename: str) -> bool:
        if self.regex:
            return bool(re.match(self.source, filename))
        else:
            return self.source == filename


class SectionConfig(BaseModel):
    name: str
    plots: list[PlotConfig]


class Config(BaseModel):
    sections: list[SectionConfig]

    @classmethod
    def from_file(cls, filename: str):
        with open(filename, "r") as handle:
            base = json.load(handle)

        return cls(
            sections=[
                SectionConfig(
                    name=key,
                    plots=[PlotConfig(source=k, **v) for k, v in value.items()],
                )
                for key, value in base.items()
            ]
        )


class Plot(BaseModel):
    filename: str
    title: str = ""
    caption: str = ""
    hash: str = ""

    def model_post_init(self, __context):
        self.hash = str(abs(hash(self.title + self.caption + self.filename)))


class Section(BaseModel):
    name: str
    plots: list[Plot]
    hash: str = ""

    def model_post_init(self, __context):
        self.hash = str(abs(hash(self.name)))


class PlotContainer(BaseModel):
    sections: list[Section]

    @classmethod
    def from_config(cls, config: Config, files: set[str]):
        sections = []
        matched_files = set()
        for section_config in config.sections:
            section_plots = []
            for plot_config in section_config.plots:
                for filename in files:
                    if plot_config.match(filename):
                        section_plots.append(
                            Plot(
                                filename=filename,
                                title=plot_config.title,
                                caption=plot_config.caption,
                            )
                        )
                        matched_files.add(filename)

            if len(section_plots) > 0:
                sections.append(Section(name=section_config.name, plots=section_plots))

        # If there are leftover files, add them to the uncategorised section
        left_over_files = matched_files ^ files

        if len(left_over_files) > 0:
            sections.append(
                Section(
                    name="Uncategorised",
                    plots=[
                        Plot(
                            filename=filename,
                        )
                        for filename in left_over_files
                    ],
                )
            )

        return PlotContainer(sections=sections)



"""
An example use of lmsi. Puts three plots into a html webpage.
"""

import matplotlib.pyplot as plt

COLORS_USED = 0


def make_figure(title: str, filename: str):
    global COLORS_USED
    fig, ax = plt.subplots()
    ax.plot([1, 2, 3], [1, 2, 3], color=f"C{COLORS_USED}")
    COLORS_USED += 1
    ax.set_title(title)
    fig.savefig(filename)
    plt.close(fig)


make_figure("Find me with regex", "12341_plot.png")
make_figure("Find me based on a filename", "no_regex_required.png")
make_figure("Find me without a caption", "no_caption.png")

# Run the script!
print("You need to run:")
print("lmsi --files *.png --output index.html --config example.json")



"""
Core functions for the LMSI package, taking in
raw paths and saving out the HTML file.
"""

from pathlib import Path

from lmsi.container import Config, PlotContainer
from lmsi.webpage.html import WebpageCreator


def core(
    potential_plots: list[Path],
    config_location: Path,
    output: Path,
):
    if config_location is not None:
        config = Config.from_file(config_location)
    else:
        config = Config(sections=[])

    plot_container = PlotContainer.from_config(
        config, {x.name for x in potential_plots}
    )

    webpage_creator = WebpageCreator()
    webpage_creator.add_metadata("LMSI Webpage")
    webpage_creator.add_plots(plot_container)
    webpage_creator.render_webpage()
    webpage_creator.save_html(output)

    return
